import pathlib
import typing as t
from dataclasses import dataclass

from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.dag.manager import DAGRunConcurrentManager
from ml_pipeline_engine.node.retrying import NodeRetryPolicy
from ml_pipeline_engine.parallelism import process_pool_registry
from ml_pipeline_engine.parallelism import threads_pool_registry
from ml_pipeline_engine.types import DAGLike
from ml_pipeline_engine.types import DAGRunManagerLike
from ml_pipeline_engine.types import NodeId
from ml_pipeline_engine.types import NodeLike
from ml_pipeline_engine.types import NodeResultT
from ml_pipeline_engine.types import PipelineContextLike
from ml_pipeline_engine.types import RetryPolicyLike


@dataclass()
class DAG(DAGLike):
    graph: DiGraph
    input_node: NodeId
    output_node: NodeId
    is_process_pool_needed: bool
    is_thread_pool_needed: bool
    node_map: t.Dict[NodeId, NodeLike]
    retry_policy: t.Type[RetryPolicyLike] = NodeRetryPolicy
    run_manager: t.Type[DAGRunManagerLike] = DAGRunConcurrentManager

    def _start_runtime_validation(self) -> None:
        self._validate_pool_executors()

    def _validate_pool_executors(self) -> None:
        if self.is_thread_pool_needed:
            threads_pool_registry.is_ready()

        if self.is_process_pool_needed:
            process_pool_registry.is_ready()

    async def run(self, ctx: PipelineContextLike) -> NodeResultT:
        self._start_runtime_validation()

        run_manager = self.run_manager(dag=self, ctx=ctx)
        return await run_manager.run()

    def visualize(  # type: ignore
        self,
        name: str,
        verbose_name: t.Optional[str] = None,
        target_dir: t.Optional[t.Union[pathlib.Path, str]] = None,
        **kwargs: t.Any,
    ) -> None:
        """
        Create static files for graph visualization

        Args:
            name: Tech name for the dag
            verbose_name: Dag title
            target_dir: Target dir for static
            **kwargs: Graph config kwargs
        """

        from ml_pipeline_engine.visualization.dag import GraphConfigImpl
        from ml_pipeline_engine.visualization.dag import build_static

        config = GraphConfigImpl(self).generate(
            name=name or 'Dag',
            verbose_name=verbose_name,
            **kwargs,
        )

        build_static(
            config,
            target_dir=pathlib.Path(target_dir) or pathlib.Path(__file__).resolve(),
        )



import typing as t

from ml_pipeline_engine.artifact_store.store.no_op import NoOpArtifactStore
from ml_pipeline_engine.events import EventSourceMixin
from ml_pipeline_engine.module_loading import get_instance
from ml_pipeline_engine.node import generate_pipeline_id
from ml_pipeline_engine.types import ArtifactStoreLike
from ml_pipeline_engine.types import EventManagerLike
from ml_pipeline_engine.types import ModelName
from ml_pipeline_engine.types import NodeId
from ml_pipeline_engine.types import PipelineChartLike
from ml_pipeline_engine.types import PipelineId


class DAGPipelineContext(EventSourceMixin):
    """
    Контекст выполнения пайплайна ML-модели
    """

    def __init__(
        self,
        chart: PipelineChartLike,
        pipeline_id: PipelineId = None,
        input_kwargs: t.Optional[t.Dict[str, t.Any]] = None,
        meta: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> None:
        self.chart = chart
        self.pipeline_id = pipeline_id if pipeline_id is not None else generate_pipeline_id()
        self.input_kwargs = input_kwargs if input_kwargs is not None else {}
        self.meta = meta if meta is not None else {}
        self.artifact_store: ArtifactStoreLike = get_instance(
            cls=self.chart.artifact_store or NoOpArtifactStore, ctx=self,
        )

        self._event_managers = [get_instance(cls) for cls in self.chart.event_managers]

    async def save_node_result(self, node_id: NodeId, data: t.Any) -> None:
        await self.artifact_store.save(node_id=node_id, data=data)

    @property
    def model_name(self) -> ModelName:
        return self.chart.model_name

    def _get_event_managers(self) -> t.List[EventManagerLike]:
        return self._event_managers

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} model_name="{self.chart.model_name}" pipeline_id="{self.pipeline_id}">'


def create_context_from_chart(
    chart: PipelineChartLike,
    input_kwargs: t.Dict[str, t.Any],
    pipeline_id: PipelineId = None,
    meta: t.Optional[t.Dict[str, t.Any]] = None,
) -> DAGPipelineContext:
    """
    Создать контекст выполнения пайплайна ML-модели
    """

    return DAGPipelineContext(
        pipeline_id=pipeline_id, chart=chart, input_kwargs=input_kwargs, meta=meta,
    )



import enum


class NodeType(str, enum.Enum):
    datasource = 'datasource'
    feature = 'feature'
    ml_model = 'ml_model'
    processor = 'processor'
    vectorizer = 'vectorizer'
    generic = 'generic'
    switch = 'switch'
    input_one_of = 'input_one_of'

    @classmethod
    def is_generic(cls, value: str) -> bool:
        return cls.generic.value in value.lower()

    @classmethod
    def by_prefix(cls, value: str) -> 'NodeType':

        for item in cls:
            if value.startswith(item):
                return cls(item)

        raise RuntimeError("Couldn't find the type")


class NodeTag(str, enum.Enum):
    """
    Поддерживаемые теги узлов
    """

    process = 'process'
    thread = 'thread'
    non_async = 'non_async'



from distutils.dir_util import copy_tree

import importlib_resources


def copy_resources(source_package: str, *resource_path: str, target_dir: str) -> None:
    """
    Copy resources from a Python package to a target directory.

    Args:
        source_package: The name of the Python package containing the resources.
        resource_path: The path to the resources within the package.
        target_dir: The directory where the resources will be copied.
    """

    with importlib_resources.path(source_package, *resource_path) as resource_dir:
        copy_tree(resource_dir, target_dir)



import typing as t
from collections import UserDict
from dataclasses import dataclass
from dataclasses import field

from ml_pipeline_engine.types import NodeId


class HiddenDict(UserDict):
    """
    Dict object that can hide some keys until they are set again
    """

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._hidden_keys: set = set()

    def get(self, key: t.Any, with_hidden: bool = True) -> t.Any:
        if with_hidden is False and key in self._hidden_keys:
            return None

        return super().get(key)

    def exists(self, key: t.Any, with_hidden: bool = True) -> bool:
        if with_hidden is False and key in self._hidden_keys:
            return False

        return key in self

    def set(self, key: t.Any, value: t.Any) -> None:

        if key in self._hidden_keys:
            self._hidden_keys.remove(key)

        self[key] = value

    def hide(self, key: t.Any) -> None:
        self._hidden_keys.add(key)

    def delete(self, key: t.Any) -> None:
        self.pop(key)


@dataclass
class DAGNodeStorage:
    """
    A container for all information about node results
    """

    node_results: HiddenDict = field(default_factory=HiddenDict)
    processed_nodes: HiddenDict = field(default_factory=HiddenDict)
    switch_results: HiddenDict = field(default_factory=HiddenDict)
    recurrent_subgraph: HiddenDict = field(default_factory=HiddenDict)
    waiting_list: HiddenDict = field(default_factory=HiddenDict)

    def set_node_result(self, node_id: NodeId, data: t.Any) -> None:
        self.node_results.set(node_id, data)

    def get_node_result(self, node_id: NodeId, with_hidden: bool = False) -> t.Any:
        return self.node_results.get(node_id, with_hidden)

    def hide_node_result(self, node_id: NodeId) -> None:
        self.node_results.hide(node_id)

    def exists_node_result(self, node_id: NodeId, with_hidden: bool = False) -> bool:
        return self.node_results.exists(node_id, with_hidden)

    def exists_node_error(self, node_id: NodeId, with_hidden: bool = False) -> bool:
        return isinstance(self.node_results.get(node_id, with_hidden), BaseException)

    def set_switch_result(self, node_id: NodeId, data: t.Any) -> t.Any:
        self.switch_results.set(node_id, data)

    def get_switch_result(self, node_id: NodeId, with_hidden: bool = False) -> t.Any:
        return self.switch_results.get(node_id, with_hidden)

    def set_node_as_processed(self, node_id: NodeId) -> None:
        self.processed_nodes.set(node_id, 1)

    def hide_processed_node(self, node_id: NodeId) -> None:
        self.processed_nodes.hide(node_id)

    def exists_processed_node(self, node_id: NodeId, with_hidden: bool = False) -> bool:
        return self.processed_nodes.exists(node_id, with_hidden)

    def set_active_rec_subgraph(self, source: NodeId, dest: NodeId) -> None:
        self.processed_nodes.set((source, dest), 1)

    def delete_active_rec_subgraph(self, source: NodeId, dest: NodeId) -> None:
        self.processed_nodes.delete((source, dest))

    def exists_active_rec_subgraph(self, source: NodeId, dest: NodeId) -> bool:
        return self.processed_nodes.exists((source, dest))

    def hide_last_execution(self, *node_ids: NodeId) -> None:
        for node_id in node_ids:
            self.hide_processed_node(node_id)
            self.hide_node_result(node_id)



from enum import Enum


class NodeField(str, Enum):
    is_switch = 'is_switch'
    is_oneof_head = 'is_oneof'
    is_oneof_child = 'is_oneof_child'
    oneof_nodes = 'oneof_nodes'
    start_node = 'start_node'
    max_iterations = 'max_iterations'
    additional_data = 'additional_data'


class EdgeField(str, Enum):
    kwarg_name = 'kwarg_name'
    is_switch = 'is_switch'
    case_branch = 'case_branch'



import logging

logger_manager = logging.getLogger('pipeline_engine.manager')
logger_manager_lock = logging.getLogger('pipeline_engine.manager.locks')
logger_node = logging.getLogger('pipeline_engine.node')
logger_decorators = logging.getLogger('pipeline_engine.decorators')
logger_parallelism = logging.getLogger('pipeline_engine.parallelism')
logger_visualization = logging.getLogger('pipeline_engine.visualization')



import itertools
import typing as t

import networkx as nx

from ml_pipeline_engine.types import NodeId


class DiGraph(nx.DiGraph):

    def __init__(self, is_recurrent: bool = False, is_oneof: bool = False, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)

        self.is_recurrent = is_recurrent
        self.is_oneof = is_oneof

    def __hash__(self) -> int:
        return hash(tuple(sorted(itertools.chain(*self.nodes.keys(), *self.edges.keys()))))


def get_connected_subgraph(
    dag: nx.Graph,
    source: NodeId,
    dest: NodeId,
    is_recurrent: bool = False,
    is_oneof: bool = False,
) -> DiGraph:
    """
    Get a connected subgraph between two nodes
    """

    if len(dag) == 1:
        return t.cast(DiGraph, dag)

    subgraph: DiGraph = dag.subgraph({node_id for path in nx.all_simple_paths(dag, source, dest) for node_id in path})
    subgraph.is_recurrent = is_recurrent
    subgraph.is_oneof = is_oneof
    subgraph.name = f'{source} —> {dest}, rec={is_recurrent}, oneof={is_oneof}'

    return subgraph



import inspect
import json
import pathlib
import typing as t
import warnings

from ml_pipeline_engine import const
from ml_pipeline_engine.node import get_callable_run_method
from ml_pipeline_engine.node.enums import NodeType
from ml_pipeline_engine.types import DAGLike
from ml_pipeline_engine.types import NodeId
from ml_pipeline_engine.types import NodeLike
from ml_pipeline_engine.visualization import schema
from ml_pipeline_engine.visualization.utils import copy_resources

_HexColorT = t.TypeVar('_HexColorT', bound=str)
_NodeTypeT = t.TypeVar('_NodeTypeT', bound=t.Union[str, NodeType])
_NodeColorsT = t.Dict[_NodeTypeT, _HexColorT]


class GraphConfigImpl:

    def __init__(self, dag: DAGLike) -> None:
        self._dag = dag

    def _get_node(self, node_id: NodeId) -> t.Type[NodeLike]:
        """
        Get a node object. Sometimes it can be None if we work with an artificial node
        """
        return self._dag.node_map.get(node_id)

    @staticmethod
    def _get_node_relative_path(node: t.Type[NodeLike]) -> str:
        """
        Generate relative path for a node
        """

        generic_class = getattr(node, '__generic_class__', None)

        if generic_class is None:
            file_path = '/'.join(node.__module__.split('.'))

        else:
            file_path = '/'.join(generic_class.__module__.split('.'))
            node = generic_class

        line_number = inspect.getsourcelines(node)[-1]
        return f'{file_path}.py#L{line_number}'

    def _generate_nodes(self) -> t.List[schema.Node]:
        """
        Generate physical and artificial nodes
        """

        nodes = []

        for node_id in self._dag.graph.nodes:
            node = self._get_node(node_id)

            if node is None:
                nodes.append(
                    schema.Node(
                        id=node_id,
                        is_virtual=True,
                        is_generic=False,
                        type=NodeType.by_prefix(node_id).value,
                    ),
                )

            else:
                method = get_callable_run_method(node)

                nodes.append(
                    schema.Node(
                        id=node_id,
                        type=node.node_type,
                        is_virtual=False,
                        is_generic=NodeType.is_generic(node.__name__),
                        data=schema.NodeAttributes(
                            name=node.name,
                            verbose_name=node.verbose_name,
                            doc=inspect.getdoc(method) or inspect.getdoc(node),
                            code_source=self._get_node_relative_path(node),
                        ),
                    ),
                )

        return nodes

    def _generate_edges(self) -> t.List[schema.Edge]:
        """
        Generate edges between physical and artificial nodes
        """
        return [
            schema.Edge(source=source, target=target)
            for source, target in self._dag.graph.edges
        ]

    def _generate_node_types(self, node_colors: t.Optional[_NodeColorsT] = None) -> t.Dict[str, schema.NodeType]:
        """
        Generate all node types.
        Will skip nodes without type.
        """
        node_colors = node_colors or {}
        node_types = {}

        for node_id in self._dag.graph.nodes:
            node = self._get_node(node_id)

            if node is None:
                node_type = NodeType.by_prefix(node_id)

            elif node.node_type is None:
                node_type = None
                warnings.warn(f'Node {node_id} without node type.', stacklevel=1)

            else:
                node_type = NodeType(node.node_type)

            if node_type in node_types or node_type is None:
                continue

            node_types[node_type.value] = schema.NodeType(
                name=node_type.value,
                hex_bgr_color=node_colors.get(node_type.value),
            )

        return node_types

    def generate(
        self,
        name: str,
        verbose_name: t.Optional[str] = None,
        repo_link: t.Optional[str] = None,
        node_colors: t.Optional[_NodeColorsT] = None,
        **kwargs: t.Any,
    ) -> schema.GraphConfig:
        """
        Generate a config for graph visualizer

        Args:
            name: Tech name for the graph
            verbose_name: Name for the graph
            repo_link: Repo link with commit or without
            node_colors: Mapping of node type colors
            **kwargs: Attrs for the graph
        """

        return schema.GraphConfig(
            nodes=self._generate_nodes(),
            edges=self._generate_edges(),
            node_types=self._generate_node_types(node_colors),
            attributes=schema.GraphAttributes(
                verbose_name=verbose_name or name,
                repo_link=repo_link,
                name=name,
                **kwargs,
            ),
        )


def build_static(config: schema.GraphConfig, target_dir: pathlib.Path) -> None:
    """
    Generate static using config and place it in the target dir
    """

    copy_resources(
        const.LIB_ANCHOR, 'visualization', 'viewer',
        target_dir=str(target_dir),
    )

    config = json.dumps(config.as_dict(), ensure_ascii=False, indent=2)

    with pathlib.Path(target_dir / 'data.js').open('w', encoding='utf-8') as file:
        file.write(f'window.__GRAPH_DATA__ = {config}')



import typing as t

from ml_pipeline_engine.node.enums import NodeType
from ml_pipeline_engine.types import NodeBase


class FeatureVectorizerBase(NodeBase):
    """
    Базовый класс для векторизаторов
    """

    node_type = NodeType.vectorizer.value

    def vectorize(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        raise NotImplementedError('Method vectorize() is not implemented')



import pathlib
from typing import List
from typing import Optional
from typing import Tuple

import click

from ml_pipeline_engine.node.enums import NodeType


@click.group()
def main() -> None:
    """Cli"""


@main.command()
@click.option('--dag_path', help='Dotted path to a dag', required=True)
@click.option('--dag_name', default='Dag', help='Tech name for the dag')
@click.option('--dag_verbose_name', default='Dag', help='Title for the dag')
@click.option('--target_dir', default='public', help='Target fir for static', type=pathlib.Path)
@click.option('--repo_link', help='Link for your repository')
@click.option('--color', type=(NodeType, str), multiple=True, help='Color for node types')
def build_static(
    dag_path: str,
    dag_name: str,
    dag_verbose_name: str,
    target_dir: pathlib.Path,
    repo_link: Optional[str] = None,
    color: Optional[List[Tuple[NodeType, str]]] = None,
) -> None:
    """Build static for a dag by path"""

    import importlib

    from ml_pipeline_engine.visualization.dag import GraphConfigImpl
    from ml_pipeline_engine.visualization.dag import build_static

    module, dag_object_name = dag_path.rsplit(':', 1)
    module = importlib.import_module(module)

    config = GraphConfigImpl(getattr(module, dag_object_name)).generate(
        node_colors=dict(color),
        repo_link=repo_link,
        verbose_name=dag_verbose_name,
        name=dag_name,
    )

    build_static(config, target_dir=target_dir)


if __name__ == '__main__':
    main()



import asyncio
import functools
import typing as t
from collections import defaultdict
from contextlib import suppress
from dataclasses import dataclass
from dataclasses import field

import networkx as nx
from cachetools import cachedmethod
from cachetools.keys import hashkey

from ml_pipeline_engine.dag.enums import EdgeField
from ml_pipeline_engine.dag.enums import NodeField
from ml_pipeline_engine.dag.errors import OneOfDoesNotHaveResultError
from ml_pipeline_engine.dag.errors import OneOfSubgraphDagError
from ml_pipeline_engine.dag.errors import RecurrentSubgraphDoesNotHaveResultError
from ml_pipeline_engine.dag.graph import DiGraph
from ml_pipeline_engine.dag.graph import get_connected_subgraph
from ml_pipeline_engine.dag.storage import DAGNodeStorage
from ml_pipeline_engine.logs import logger_manager as logger
from ml_pipeline_engine.logs import logger_manager_lock as lock_logger
from ml_pipeline_engine.node import run_node
from ml_pipeline_engine.node import run_node_default
from ml_pipeline_engine.node.retrying import NodeRetryPolicy
from ml_pipeline_engine.types import CaseResult
from ml_pipeline_engine.types import DAGLike
from ml_pipeline_engine.types import DAGRunManagerLike
from ml_pipeline_engine.types import NodeId
from ml_pipeline_engine.types import NodeResultT
from ml_pipeline_engine.types import PipelineContextLike
from ml_pipeline_engine.types import Recurrent

_EventDictT = t.Dict[t.Any, asyncio.Event]
_EventConditionT = t.Dict[t.Any, asyncio.Condition]


@dataclass
class DAGConcurrentManagerLock:
    node_ids: t.Iterable[NodeId]
    event_lock_store: _EventDictT = field(
        default_factory=functools.partial(defaultdict, asyncio.Event),
    )
    condition_lock_store: _EventConditionT = field(
        default_factory=functools.partial(defaultdict, asyncio.Condition),
    )

    @property
    def conditions(self) -> _EventConditionT:
        return self.condition_lock_store

    @property
    def events(self) -> _EventDictT:
        return self.event_lock_store

    async def wait_for_condition(self, condition_name: t.Any, condition: t.Callable) -> None:
        cond = self.conditions[condition_name]

        async with cond:
            lock_logger.debug('Lock %s', condition_name)
            await cond.wait_for(condition)

        lock_logger.debug('Unlock %s', condition_name)

    async def unlock_condition(self, condition_name: t.Any) -> None:
        """
        Send a notification to waiters
        """

        condition = self.conditions[condition_name]

        async with condition:
            lock_logger.debug('Send a notification to unlock %s', condition_name)
            condition.notify_all()


def cache_key(prefix: str, _: t.Any, *args: t.Any, **kwargs: t.Any) -> t.Type[tuple]:
    """Custom func key generation excluding 'self'."""
    return hashkey(*args, prefix, **kwargs)


@dataclass
class DAGRunConcurrentManager(DAGRunManagerLike):
    """
    Дефолтный менеджер запуска графов.
    Производит конкурентное исполнение узлов.
    """

    ctx: PipelineContextLike
    dag: DAGLike

    _node_storage: DAGNodeStorage = field(default_factory=DAGNodeStorage)
    _lock_manager: DAGConcurrentManagerLock = field(init=False)
    _memorization_store: t.Dict[t.Any, t.Any] = field(default_factory=dict)
    _coro_tasks: t.Set[asyncio.Task] = field(default_factory=set)
    _run_method_alias: str = 'run'

    def __post_init__(self) -> None:
        self._lock_manager = DAGConcurrentManagerLock(self.dag.node_map.keys())

    @staticmethod
    def _stop_coro_tasks(coro_tasks: t.Iterable[asyncio.Task]) -> None:
        """
        Stop running the coro tasks
        """

        for coro_task in coro_tasks:
            if coro_task.done():
                continue

            coro_task.cancel()
            logger.debug('Task %s has been cancelled', coro_task.get_name())

    @staticmethod
    def _get_first_error_in_tasks(coro_tasks: t.Iterable[asyncio.Task]) -> t.Optional[t.Type[Exception]]:
        """
        Check if there is an error in the coro tasks and return the first one
        """

        for coro_task in coro_tasks:
            if coro_task.done() and isinstance(coro_task.exception(), BaseException):
                return coro_task.exception()

        return None

    def _create_task(self, coro: t.Coroutine, name: str) -> asyncio.Task:
        """
        Create asyncio.Task and collect it to the main DAG's storage
        """

        task = asyncio.create_task(coro, name=name)
        self._coro_tasks.add(task)

        return task

    async def run(self) -> NodeResultT:
        """
        Run the main DAG
        """

        try:
            self._create_task(
                self._run_dag(
                    self._get_reduced_dag(self.dag.input_node, self.dag.output_node),
                    is_under_oneof=False,
                ),
                self._run_method_alias,
            )

            await self._lock_manager.wait_for_condition(
                self._run_method_alias,
                self._has_dag_result,
            )

            return self._get_dag_result()

        finally:
            self._stop_coro_tasks(self._coro_tasks)

    def _has_dag_result(self) -> bool:
        """
        Check if the dag has a result or an error
        """

        return (
            bool(self._get_first_error_in_tasks(self._coro_tasks))
            or self._node_storage.exists_node_result(self.dag.output_node)
        )

    def _get_dag_result(self) -> NodeResultT:
        """
        Get the DAG's result or raise an error if there are any errors
        """

        logger.debug('Getting the result of the dag')

        error = self._get_first_error_in_tasks(self._coro_tasks)
        if error:
            raise error

        return self._node_storage.get_node_result(self.dag.output_node, with_hidden=True)

    def _get_node_kwargs(self, node_id: NodeId) -> t.Dict[str, t.Any]:
        """
        Get the node's dependencies that are needed to run the node
        """

        kwargs = {}

        if node_id != self.dag.input_node:
            for pred_node_id in self.dag.graph.predecessors(node_id):
                kwarg_name = self.dag.graph.edges[(pred_node_id, node_id)].get(EdgeField.kwarg_name)

                if kwarg_name is None:
                    continue

                if self._is_switch(pred_node_id):
                    kwargs[kwarg_name] = self._node_storage.get_node_result(
                        self._node_storage.get_switch_result(pred_node_id).node_id,
                    )

                else:
                    kwargs[kwarg_name] = self._node_storage.get_node_result(
                        pred_node_id,
                    )

        else:
            kwargs = self.ctx.input_kwargs

        additional_data = self.dag.graph.nodes[node_id].get(NodeField.additional_data)

        if additional_data is not None:
            kwargs[NodeField.additional_data] = additional_data

        return kwargs

    def _is_switch(self, node_id: NodeId) -> bool:
        """
        Checks if the node is a switch
        """

        try:
            return self.dag.graph.nodes[node_id].get(NodeField.is_switch) is True
        except KeyError:
            return False

    def _is_head_of_oneof(self, node_id: NodeId) -> bool:
        """
        Checks if the node is the head of OneOf
        """
        return bool(self.dag.graph.nodes[node_id].get(NodeField.is_oneof_head))

    def _get_reduced_dag(self, source: NodeId, dest: NodeId) -> DiGraph:
        """
        Get filtered and connected subgraph
        """

        def _filter(u: str, v: str) -> bool:
            """
            Delete edges with EdgeField.case_branch from subgraph_view

            Args:
                u - Node
                v - Node Edge
            """
            return not self.dag.graph.edges[u, v].get(EdgeField.case_branch)

        def _filter_node(u: str) -> bool:
            """
            Delete nodes with NodeField.is_oneof_child from subgraph_view

            Args:
                u -  Node
            """
            return not self.dag.graph.nodes[u].get(NodeField.is_oneof_child)

        return get_connected_subgraph(
            nx.subgraph_view(self.dag.graph, filter_edge=_filter, filter_node=_filter_node), source, dest,
        )

    def _get_reduced_dag_input_one_of(self, source: NodeId, dest: NodeId) -> DiGraph:
        """
        Get the subgraph for the OneOf subgraph
        """

        return get_connected_subgraph(nx.subgraph_view(self.dag.graph), source, dest, is_oneof=True)

    def _add_case_result(self, switch_node_id: NodeId) -> None:
        """
        Save the switch branch
        """

        selected_branch_label = None
        branch_nodes = {}
        for pred_id in self.dag.graph.predecessors(switch_node_id):
            edge = self.dag.graph.edges[(pred_id, switch_node_id)]

            if edge.get(EdgeField.is_switch):
                selected_branch_label = self._node_storage.get_node_result(pred_id)
                continue

            branch_nodes[edge.get(EdgeField.case_branch)] = pred_id

        self._node_storage.set_switch_result(
            switch_node_id,
            CaseResult(label=selected_branch_label, node_id=branch_nodes[selected_branch_label]),
        )

    async def _execute_node(
        self,
        node_id: NodeId,
        is_under_oneof: bool = False,
        force_default: bool = False,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Execute node and save metadata
        """

        if self._node_storage.exists_processed_node(node_id):
            logger.debug('Node %s has been executed. Stop new execution', node_id)

            event_lock = self._lock_manager.events[node_id]
            await event_lock.wait()

            return self._node_storage.get_node_result(node_id)

        self._node_storage.set_node_as_processed(node_id)
        await self.ctx.emit_on_node_start(node_id=node_id)

        try:
            logger.info('Preparing node for the execution, node_id=%s', node_id)

            result = await self.__execute_node(
                node_id=node_id,
                force_default=force_default,
                **self._get_node_kwargs(node_id),
            )

            await self.ctx.emit_on_node_complete(node_id=node_id, error=None)

            logger.info('Getting the result after the execution, node_id=%s', node_id)
            return result

        except Exception as ex:
            await self.ctx.emit_on_node_complete(node_id=node_id, error=ex)
            logger.error('Execution error node_id=%s', node_id, exc_info=ex)

            if is_under_oneof:
                return ex

            raise ex

    async def __execute_node(
        self,
        node_id: NodeId,
        force_default: bool = False,
        **kwargs: t.Any,
    ) -> t.Union[NodeResultT, t.Any]:
        """
        Execute the node by node id with retry protocol

        Args:
            node_id: Node id
            force_default: If the node should return default result
            **kwargs: Key value args for the node
        """

        node = self.dag.node_map[node_id]

        retry_policy = NodeRetryPolicy(node=node)

        n_attempts = 1
        while True:
            try:
                if force_default:
                    return run_node_default(node, **kwargs)

                logger.debug('Start execution node_id=%s', node_id)
                result = await run_node(**kwargs, node=node, node_id=node_id)

                logger.debug('Finish the node execution, node_id=%s', node_id)
                return result

            except retry_policy.exceptions as error:  # noqa: PERF203
                logger.debug(
                    'Node %s will be restarted in %s seconds...',
                    node_id,
                    retry_policy.delay,
                    exc_info=error,
                )

                if n_attempts == retry_policy.attempts:

                    if node.use_default:
                        return run_node_default(node, **kwargs)

                    raise error

                await self.ctx.emit_on_node_complete(node_id=node_id, error=error)

                n_attempts += 1
                await asyncio.sleep(retry_policy.delay)

            except Exception:
                if node.use_default:
                    return run_node_default(node, **kwargs)

                raise

    def _get_node_order(self, dag: DiGraph) -> t.List[NodeId]:
        """
        Calculate the order for nodes according to the dag's type
        """

        return [
            node_id for node_id in nx.topological_sort(dag)
            if (
                not self._node_storage.exists_processed_node(node_id)
                if not dag.is_recurrent
                else True
            )
        ]

    @cachedmethod(lambda self: self._memorization_store, key=functools.partial(cache_key, 'node_dependencies'))
    def _get_node_dependencies(self, dag: DiGraph, node_id: NodeId) -> t.Set[NodeId]:
        """
        Get the node's dependencies
        """

        node_predecessors = set(self.dag.graph.predecessors(node_id))
        current_dag = set(nx.topological_sort(dag))

        return current_dag.intersection(node_predecessors)

    def _get_predecessors(self, dag: DiGraph, node_id: NodeId) -> t.List[NodeId]:
        """
        Get the node's predecessors
        """

        predecessors = list(
            self._get_node_dependencies(dag, node_id)
            if self._is_switch(node_id) or self._is_head_of_oneof(node_id) or dag.is_recurrent
            else self.dag.graph.predecessors(node_id),
        )

        for idx, node_id in enumerate(predecessors):

            if self._is_switch(node_id):
                with suppress(KeyError, AttributeError):
                    predecessors[idx] = self._node_storage.get_switch_result(node_id).node_id

        return predecessors

    def _is_ready_to_execute(self, dag: DiGraph, node_id: NodeId) -> bool:
        """
        Check if the node is read to be executed
        """

        logger.debug('Checking if the node can be executed node_id=%s', node_id)

        for pred_node_id in self._get_predecessors(dag, node_id):

            if (
                not self._node_storage.exists_node_result(pred_node_id)
                # The node cannot be executed if there is a "Recurrent" result in the node's dependencies.
                # Hence, the node should wait for proper a result or an error.
                or isinstance(self._node_storage.get_node_result(pred_node_id), Recurrent)
            ):
                logger.debug(
                    'The node %s cannot be executed due to absense the dependent result of the node %s',
                    node_id,
                    pred_node_id,
                )

                return False

        logger.debug('The node can be executed node_id=%s', node_id)
        return True

    async def _run_dag(self, dag: DiGraph, is_under_oneof: bool, dag_output_node: t.Optional[NodeId] = None) -> t.Any:
        """
        Run the dag.
        """

        logger.debug('Start DAG execution, dag=%s', str(dag))

        list_node_ids = self._get_node_order(dag)

        if dag.is_recurrent:
            logger.debug('Hide previous node results for recurrent subgraph %s', list_node_ids)
            self._node_storage.hide_last_execution(*list_node_ids)

        if len(list_node_ids) == 0:
            return None

        dag_output_node = dag_output_node if dag_output_node is not None else list_node_ids[-1]
        last_node = list_node_ids[-1]
        local_tasks = []

        for node_id in list_node_ids:

            await self._lock_manager.wait_for_condition(
                node_id,
                functools.partial(self._is_ready_to_execute, dag, node_id),
            )

            if is_under_oneof:
                for local_task in local_tasks:

                    # If we have an error in previous nodes, we don't need to continue running OneOf subgraph.
                    # It means the next subgraph in OneOf will be executed.
                    if self._node_storage.exists_node_error(local_task.get_name()):

                        self._stop_coro_tasks(local_tasks)
                        raise OneOfSubgraphDagError

            if self._is_switch(node_id):
                coro_to_run = self._run_switch(node_id, is_under_oneof)

            elif self._is_head_of_oneof(node_id):
                coro_to_run = self._run_oneof(node_id)

            else:
                coro_to_run = self._run_node(
                    node_id,
                    is_under_oneof=is_under_oneof,
                    output_node_id=dag_output_node,
                    is_last_node=node_id == last_node,
                )

            local_tasks.append(self._create_task(coro_to_run, name=node_id))

        await self._lock_manager.wait_for_condition(
            dag_output_node,
            functools.partial(self._node_storage.exists_node_result, dag_output_node),
        )

        return self._node_storage.get_node_result(dag_output_node, with_hidden=True)

    async def _run_oneof(self, node_id: NodeId) -> t.Any:
        """
        Run OneOf subgraph. Returns the first non-error result or specific error
        """

        logger.debug('Prepare OneOf DAG node_id=%s', node_id)

        for subgraph_node_id in self.dag.graph.nodes[node_id][NodeField.oneof_nodes]:
            logger.debug(
                'Prepare OneOf subgraph to start target_node_id=%s, subgraph_node_id=%s',
                node_id,
                subgraph_node_id,
            )

            with suppress(OneOfSubgraphDagError):

                return await self._run_dag(
                    dag_output_node=node_id,
                    is_under_oneof=True,
                    dag=self._get_reduced_dag_input_one_of(
                        self.dag.input_node,
                        subgraph_node_id,
                    ),
                )

        raise OneOfDoesNotHaveResultError(node_id)

    async def _run_switch(self, node_id: NodeId, is_under_oneof: bool) -> t.Any:
        """
        Run switch subgraph
        """

        logger.debug('Prepare Switch DAG node_id=%s', node_id)

        self._add_case_result(node_id)

        return await self._run_dag(
            is_under_oneof=is_under_oneof,
            dag=self._get_reduced_dag(
                self.dag.input_node,
                (self._node_storage.get_switch_result(node_id)).node_id,
            ),
        )

    async def _run_node(
        self,
        node_id: NodeId,
        output_node_id: NodeId,
        is_under_oneof: bool = False,
        is_last_node: bool = False,
        force_default: bool = False,
    ) -> None:
        """
        Method runs the node and according to the execution result orchestrates the node's locks
        in order to unlock all dependencies and itself

        Args:
            node_id: Node id to execute
            output_node_id: Output Node id
            force_default: If the node should be executed with default result
            is_under_oneof: If the node is performing under the OneOf subgraph
            is_last_node: If the node is the last node in its DAG
        """

        to_unlock_descendants = True
        node_ids = (node_id,)

        try:
            result = await self._execute_node(
                node_id,
                is_under_oneof=is_under_oneof,
                force_default=force_default,
            )

            if isinstance(result, Recurrent):
                self._create_task(
                    self._run_recurrent_subgraph(
                        node_id, node_result=result, is_under_oneof=is_under_oneof,
                    ),
                    f'recurrent-subgraph-dest-{node_id}',
                )

                # We shouldn't unlock the node's descendants if we have to perform recurrent subgraph.
                # It has to be this way because the node, which has `Recurrent` result,
                # will be executed again and the function will unlock the descendants in the other branch.
                to_unlock_descendants = False

            elif is_under_oneof and is_last_node and not isinstance(result, BaseException):
                node_ids = (node_id, output_node_id)

            for node in node_ids:
                logger.debug('Save the result "%s" for the node %s', result, node)
                self._node_storage.set_node_result(node, result)

                # TODO: Needs to reorganize saving policy for artifact storage
                await self.ctx.save_node_result(node, result)

        finally:
            if not to_unlock_descendants:
                logger.debug('Skip unlocking the descendants of the node, node_id=%s', node_id)
                self.__unlock_execution_lock(node_id)

                # Unlock itself to perform the next step in the node's DAG
                await self._lock_manager.unlock_condition(node_id)

                return  # noqa: B012

            for node in node_ids:
                logger.debug('Start the procedure of unlocking dependencies node_id=%s', node)

                self.__unlock_execution_lock(node)

                await self.__unlock_descendants(node)
                await self.__unlock_run_method()

                if node == output_node_id:
                    logger.debug('The node %s is an output node', node)
                    await self._lock_manager.unlock_condition(node)

    async def __unlock_run_method(self) -> None:
        """
        Unlock the main method in order to return the main DAG result
        """

        await self._lock_manager.unlock_condition(self._run_method_alias)

    def __unlock_execution_lock(self, node_id: NodeId) -> None:
        """
        Unlock the execution function for the node that can be executed concurrently
        """

        lock_logger.debug('Unlock execution for %s if there is any concurrent execution', node_id)
        self._lock_manager.events[node_id].set()

    async def _run_recurrent_subgraph(self, node_id: NodeId, node_result: Recurrent, is_under_oneof: bool) -> None:
        """
        Run a recurrent subgraph with

        Args:
            node_id: Node id as the end of the subgraph
            node_result: Previous subgraph's result
            is_under_oneof: If the subgraph is performing under the OneOf subgraph
        """

        start_from_node_id = self.dag.graph.nodes[node_id].get(NodeField.start_node)

        if self._node_storage.exists_active_rec_subgraph(start_from_node_id, node_id):
            return

        self._node_storage.set_active_rec_subgraph(start_from_node_id, node_id)
        max_iterations = self.dag.graph.nodes[node_id].get(NodeField.max_iterations)

        logger.debug(
            'Start the process of the recurrent subgraph for nodes start_node=%s, dest_node=%s',
            start_from_node_id,
            node_id,
        )

        recurrent_subgraph = get_connected_subgraph(
            self.dag.graph, start_from_node_id, node_id, is_recurrent=True,
        )

        for current_iter in range(max_iterations):
            logger.debug(
                'Executing the %s attempt of the recurrent subgraph start_node=%s, dest_node=%s',
                current_iter,
                start_from_node_id,
                node_id,
            )
            start_node = self.dag.graph.nodes[start_from_node_id]
            start_node[NodeField.additional_data] = node_result.data

            node_result = await self._run_dag(dag=recurrent_subgraph, is_under_oneof=is_under_oneof)

            logger.debug(
                'Getting the result of %s attempt of the recurrent subgraph start_node=%s, dest_node=%s',
                current_iter,
                start_from_node_id,
                node_id,
            )

            if not isinstance(node_result, Recurrent):
                break

        else:
            node = self.dag.node_map[node_id]

            if isinstance(node_result, Recurrent) and node.use_default:
                logger.debug(
                    'Attempts to run a recurrent subgraph have been exceeded. '
                    'Will be used the default value start_node=%s, dest_node=%s',
                    start_from_node_id,
                    node_id,
                )

                self._node_storage.hide_last_execution(node_id)

                await self._run_node(
                    node_id,
                    output_node_id=node_id,
                    is_under_oneof=is_under_oneof,
                    force_default=True,
                )

            else:
                raise RecurrentSubgraphDoesNotHaveResultError(
                    start_from_node_id,
                    node_id,
                    node_result,
                )

        logger.debug(
            'Finish the recurrent subgraph for nodes start_node=%s, dest_node=%s',
            start_from_node_id,
            node_id,
        )

        self._node_storage.delete_active_rec_subgraph(start_from_node_id, node_id)

    async def __unlock_descendants(self, node_id: NodeId) -> None:
        """
        Send a notification to the node's descendants to unlock them
        """

        descendants = self.__get_descendants(node_id)

        for descendant_node_id in descendants:
            await self._lock_manager.unlock_condition(descendant_node_id)

    def __get_descendants(self, node_id: NodeId) -> t.List[NodeId]:
        """
        Get all first-line the node's descendants including artificial nodes
        """

        logger.debug('Getting descendants for the node %s', node_id)

        descendants = list(nx.descendants_at_distance(self.dag.graph, node_id, 1))

        for idx in range(len(descendants)):

            descendant_node_id = descendants[idx]

            if self._is_switch(descendant_node_id):
                descendants.extend(self.__get_descendants(descendant_node_id))

        return descendants

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} nnodes="{len(self.dag.graph.nodes)}" nedges="{len(self.dag.graph.edges)}">'



import functools
import typing as t
import warnings
from enum import Enum
from pathlib import Path

from ml_pipeline_engine.artifact_store.enums import DataFormat
from ml_pipeline_engine.artifact_store.errors import ArtifactAlreadyExists
from ml_pipeline_engine.artifact_store.errors import ArtifactDoesNotExist
from ml_pipeline_engine.artifact_store.serializers import serializer_factory
from ml_pipeline_engine.artifact_store.store.base import SerializedArtifactStore
from ml_pipeline_engine.types import NodeId
from ml_pipeline_engine.types import NodeResultT
from ml_pipeline_engine.types import PipelineContextLike


class ArtifactFileAlreadyExists(ArtifactAlreadyExists):
    pass


class ArtifactFileDoesNotExist(ArtifactDoesNotExist):
    pass


def dont_use_for_prod(func: t.Callable) -> t.Callable[..., t.Any]:

    @functools.wraps(func)
    async def wrap(*args: t.Any, **kwargs: t.Any) -> t.Any:
        warnings.warn(f'Функция {func.__name__} предназначена для локального использования', stacklevel=1)
        return await func(*args, **kwargs)

    return wrap


class FileSystemArtifactStore(SerializedArtifactStore):
    def __init__(self, ctx: PipelineContextLike, artifact_dir: t.Union[Path, str]) -> None:
        super().__init__(ctx)

        self.artifact_dir = Path(artifact_dir)

    def _ensure_dir(self) -> Path:
        model_name = self.ctx.model_name.value if isinstance(self.ctx.model_name, Enum) else self.ctx.model_name
        path = Path(self.artifact_dir / model_name / str(self.ctx.pipeline_id))

        if not path.exists():
            path.mkdir(parents=True)

        return path

    def _get_glob(self, node_id: NodeId) -> t.List[Path]:
        return list(Path(self._ensure_dir()).glob(f'{node_id}.*'))

    @dont_use_for_prod
    async def save(self, node_id: NodeId, data: NodeResultT, fmt: DataFormat = DataFormat.PICKLE) -> None:
        if len(self._get_glob(node_id)):
            raise ArtifactFileAlreadyExists(f'Artifact file for {node_id} already exists')

        with Path(self._ensure_dir() / f'{node_id}.{fmt.value}').open('wb') as file:  # noqa: ASYNC101
            serializer_factory.from_data_format(fmt).dump(data, file)

    @dont_use_for_prod
    async def load(self, node_id: NodeId) -> NodeResultT:
        glob = self._get_glob(node_id)

        if not len(glob):
            raise ArtifactFileDoesNotExist(f'Artifact file for {node_id} does not exist')

        with Path(glob[0]).open('rb') as file:  # noqa: ASYNC101
            return serializer_factory.from_extension(glob[0].suffix[1:]).load(file)



from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List
from typing import Optional


@dataclass
class GraphAttributes:
    verbose_name: str
    name: str  # Tech name
    repo_link: Optional[str] = None
    edgesep: int = field(default=60)  # Number of pixels that separate edges horizontally in the layout.
    ranksep: int = field(default=700)  # Number of pixels between each rank in the layout.


@dataclass
class NodeAttributes:
    name: str  # Tech name
    verbose_name: str
    doc: str
    code_source: str


@dataclass
class Node:
    id: str
    is_virtual: bool
    is_generic: bool
    type: Optional[str] = None  # If a node is artificial
    data: Optional[NodeAttributes] = None  # If a node is artificial


@dataclass
class Edge:
    id: str = field(init=False)
    source: str
    target: str

    def __post_init__(self) -> None:
        self.id = f'{self.source}->{self.target}'


@dataclass
class NodeType:
    name: str  # Tech name
    hex_bgr_color: Optional[str] = None


@dataclass
class GraphConfig:
    nodes: List[Node]
    edges: List[Edge]
    node_types: Dict[str, NodeType]
    attributes: GraphAttributes

    def as_dict(self) -> Dict[str, Any]:
        return asdict(self)



from ml_pipeline_engine.base_nodes.processors import ProcessorBase
from ml_pipeline_engine.dag_builders.annotation import build_dag
from ml_pipeline_engine.dag_builders.annotation.marks import Input
from ml_pipeline_engine.dag_builders.annotation.marks import SwitchCase


class Ident(ProcessorBase):
    """
    Возвращает входное значение таким, каким оно было
    """
    name = 'ident'
    verbose_name = 'Identity'

    def process(self, num: float) -> float:
        return num


class SwitchStmt(ProcessorBase):
    """
    Вычисляет условие для switch
    """
    name = 'switch_stmt'
    verbose_name = 'Условие для switch'

    def process(self, num: Input(Ident)) -> str:
        if num < 0.0:
            return 'invert'
        return 'nested_switch'


class Invert(ProcessorBase):
    """
    Инвертирует число
    """
    name = 'invert'
    verbose_name = 'Инвертор'

    def process(self, num: Input(Ident)) -> float:
        return -num


class NestedSwitchStmt(ProcessorBase):
    """
    Вычисляет условие для switch
    """
    name = 'nested_switch_stmt'
    verbose_name = 'Условие для вложенного switch'

    def process(self, num: Input(Ident)) -> str:
        if num == 1.0:
            return 'double'
        return 'triple'


class DoubleNumber(ProcessorBase):
    """
    Умножает число на 2
    """
    name = 'double_number'
    verbose_name = 'Удвоение числа'

    def process(self, num: Input(Ident)) -> float:
        return num * 2


class TripleNumber(ProcessorBase):
    """
    Умножает число на 3
    """
    name = 'triple_number'
    verbose_name = 'Умножение числа на 3'

    def process(self, num: Input(Ident)) -> float:
        return num * 3


NestedSwitchCase = SwitchCase(
    name='nested_switch',
    switch=NestedSwitchStmt,
    cases=[
        ('double', DoubleNumber),
        ('triple', TripleNumber),
    ],
)


class Add3Node(ProcessorBase):
    """
    Прибавляет число 3
    """
    name = 'add_3'
    verbose_name = 'Прибавление числа 3'

    def process(self, num: NestedSwitchCase) -> float:
        return num + 3


SomeSwitchCase = SwitchCase(
    name='main_switch',
    switch=SwitchStmt,
    cases=[
        ('invert', Invert),
        ('nested_switch', Add3Node),
    ],
)


class Result(ProcessorBase):
    """
    Возвращает результат
    """
    name = 'result'
    verbose_name = 'Результат'

    def process(self, num: SomeSwitchCase, num2: Input(Ident)) -> float:
        return num + num2


sample_dag = build_dag(input_node=Ident, output_node=Result)



import typing as t

from ml_pipeline_engine.node.enums import NodeType
from ml_pipeline_engine.types import NodeBase


class FeatureBase(NodeBase):
    """
    Базовый класс для набора фичей
    """

    node_type = NodeType.feature.value

    def extract(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        raise NotImplementedError('Method extract() is not implemented')



class BaseDagError(Exception):
    pass


class OneOfDoesNotHaveResultError(BaseDagError):
    pass


class OneOfSubgraphDagError(BaseDagError):
    pass


class RecurrentSubgraphDoesNotHaveResultError(BaseDagError):
    pass



import typing as t

from ml_pipeline_engine.node.enums import NodeType
from ml_pipeline_engine.types import AdditionalDataT
from ml_pipeline_engine.types import NodeBase
from ml_pipeline_engine.types import Recurrent
from ml_pipeline_engine.types import RecurrentProtocol


class ProcessorBase(NodeBase):
    """
    Базовый класс для обработчиков общего назначения
    """

    node_type = NodeType.processor.value

    def process(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        raise NotImplementedError('Method process() is not implemented')


class RecurrentProcessor(ProcessorBase, RecurrentProtocol):
    """
    Узел процессора, который может быть исполнен в рекуррентном подграфе
    """

    def next_iteration(self, data: AdditionalDataT) -> Recurrent:
        return Recurrent(data=data)
